@@ -30,6 +30,7 @@ b_builder *b_builder_new(size_t block_factor) {
 
     builder->total          = 0;
     builder->match          = NULL;
+    builder->options        = B_BUILDER_NONE;
     builder->lookup_service = NULL;
     builder->lookup_ctx     = NULL;
     builder->data           = NULL;
@@ -46,6 +47,16 @@ error_malloc:
     return NULL;
 }
 
+enum b_builder_options b_builder_get_options(b_builder *builder) {
+    if (builder == NULL) return B_BUILDER_NONE;
+
+    return builder->options;
+}
+
+void b_builder_set_options(b_builder *builder, enum b_builder_options options) {
+    builder->options = options;
+}
+
 b_error *b_builder_get_error(b_builder *builder) {
     if (builder == NULL) return NULL;
 
@@ -93,12 +104,10 @@ int b_builder_exclude_from_file(b_builder *builder, const char *file) {
     return lafe_exclude_from_file(&builder->match, file);
 }
 
-int b_builder_write_file(b_builder *builder, b_string *path, b_string *member_name, struct stat *st) {
+int b_builder_write_file(b_builder *builder, b_string *path, b_string *member_name, struct stat *st, int fd) {
     b_buffer *buf = builder->buf;
     b_error *err  = builder->err;
 
-    int file_fd = 0;
-
     off_t wrlen = 0;
 
     b_header *header;
@@ -113,16 +122,6 @@ int b_builder_write_file(b_builder *builder, b_string *path, b_string *member_na
         b_error_clear(err);
     }
 
-    if ((st->st_mode & S_IFMT) == S_IFREG) {
-        if ((file_fd = open(path->str, O_RDONLY)) < 0) {
-            if (err) {
-                b_error_set(err, B_ERROR_WARN, errno, "Cannot open file", path);
-            }
-
-            goto error_open;
-        }
-    }
-
     if ((header = b_header_for_file(path, member_name, st)) == NULL) {
         if (err) {
             b_error_set(err, B_ERROR_FATAL, errno, "Cannot build header for file", path);
@@ -160,6 +159,20 @@ int b_builder_write_file(b_builder *builder, b_string *path, b_string *member_na
     if (header->truncated) {
         b_string *longlink_path;
 
+        /*
+         * GNU extensions must be explicitly enabled to encode GNU LongLink
+         * headers.
+         */
+        if (!(builder->options & B_BUILDER_GNU_EXTENSIONS)) {
+            errno = ENAMETOOLONG;
+
+            if (err) {
+                b_error_set(err, B_ERROR_WARN, errno, "File name too long", member_name);
+            }
+
+            goto error_path_toolong;
+        }
+
         if ((block = b_buffer_get_block(buf, B_HEADER_SIZE, &wrlen)) == NULL) {
             goto error_get_header_block;
         }
@@ -207,8 +220,8 @@ int b_builder_write_file(b_builder *builder, b_string *path, b_string *member_na
     /*
      * Finally, end by writing the file contents.
      */
-    if (file_fd) {
-        if ((wrlen = b_file_write_contents(buf, file_fd, header->size)) < 0) {
+    if ((st->st_mode & S_IFMT) == S_IFREG && fd > 0) {
+        if ((wrlen = b_file_write_contents(buf, fd, header->size)) < 0) {
             if (err) {
                 b_error_set(err, B_ERROR_WARN, errno, "Cannot write file to archive", path);
             }
@@ -217,10 +230,6 @@ int b_builder_write_file(b_builder *builder, b_string *path, b_string *member_na
         }
 
         builder->total += wrlen;
-
-        close(file_fd);
-
-        file_fd = 0;
     }
 
     b_header_destroy(header);
@@ -231,16 +240,12 @@ error_write:
 error_longlink_path_append:
 error_longlink_path_dup:
 error_get_header_block:
+error_path_toolong:
 error_header_encode:
 error_lookup:
     b_header_destroy(header);
 
 error_header_for_file:
-    if (file_fd) {
-        close(file_fd);
-    }
-
-error_open:
     return -1;
 }
 
@@ -257,8 +262,9 @@ void b_builder_destroy(b_builder *builder) {
         builder->err = NULL;
     }
 
-    builder->total = 0;
-    builder->data  = NULL;
+    builder->options = B_BUILDER_NONE;
+    builder->total   = 0;
+    builder->data    = NULL;
 
     lafe_cleanup_exclusions(&builder->match);
 
